﻿using Dynamicweb.Ecommerce.Integration;
using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Ecommerce.LiveIntegration.Connectors;
using Dynamicweb.Ecommerce.LiveIntegration.Logging;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Ecommerce.Products;
using Dynamicweb.Security.UserManagement;
using System;
using System.Collections.Generic;
using System.Linq;

namespace Dynamicweb.Ecommerce.LiveIntegration.Products
{
    /// <summary>
    /// PriceProvider to find prices and other information for products.
    /// </summary>
    /// <seealso cref="PriceProvider" />
    public class ProductPriceProvider : PriceProvider
    {
        /// <summary>
        /// Finds the price..
        /// </summary>
        /// <param name="product">The product.</param>
        /// <param name="quantity">The quantity.</param>
        /// <param name="variantId">The variant id.</param>
        /// <param name="currency">The currency.</param>
        /// <param name="unitId">The unit id.</param>
        /// <param name="user">The user.</param>
        /// <returns>The <see cref="T:Dynamicweb.Ecommerce.Prices.PriceRaw" />.</returns>
        public override PriceRaw FindPrice(Product product, double quantity, string variantId, International.Currency currency, string unitId, User user)
        {
            if (!Global.IsIntegrationActive
              || !Settings.Instance.EnableLivePrices
              || (user == null && !Settings.Instance.LiveProductInfoForAnonymousUsers)
              || (user != null && user.IsLivePricesDisabled)
              || !Connector.IsWebServiceConnectionAvailable()
              || product == null
              || string.IsNullOrEmpty(product.Id)
              || string.IsNullOrEmpty(product.Number))
            {
                return null;
            }

            try
            {
                string productIdentifier = ProductManager.ProductProvider.GetProductIdentifier(product);

                // If Product is not in the cache, then get it from Erp
                if (!ErpResponseCache.IsProductInCache(ProductManager.ProductCacheLevel, productIdentifier))
                {
                    Dictionary<Product, double> products = new Dictionary<Product, double>();
                    string productVariantId = product.VariantId;
                    long productAutoId = product.AutoId;

                    Product changedProduct = ProductManager.ProductProvider.GetProductFromVariantComboId(product);
                    products.Add(changedProduct, quantity);
                    // After GetProductFromVariantComboId the variantId is empty or other product is returned so new identifier must be generated
                    productIdentifier = ProductManager.ProductProvider.GetProductIdentifier(changedProduct);

                    if (ProductManager.FetchProductInfos(products, User.GetCurrentExtranetUser()))
                    {
                        RestoreVariantId(product, productAutoId, productVariantId);
                        // Check if requested product was not received from response
                        if (!ErpResponseCache.IsProductInCache(ProductManager.ProductCacheLevel, productIdentifier))
                        {
                            Logger.Instance.Log(ErrorLevel.Error, $"Error receiving product info for product: {ProductManager.ProductProvider.GetProductIdentifier(product)}.");
                            return null;
                        }
                    }
                    else
                    {
                        RestoreVariantId(product, productAutoId, productVariantId);
                        // No price from ERP so read DW default price
                        var priceProvider = new DefaultPriceProvider();
                        return priceProvider.FindPrice(product, quantity, variantId, currency, unitId, user);
                    }
                }

                double? priceValue = 0.0;

                var productCache = ErpResponseCache.GetProductInfos(ProductManager.ProductCacheLevel);

                ProductInfo productInfo = productCache != null && productCache.TryGetValue(productIdentifier, out productInfo)
                    ? productInfo
                    : null;

                if (productInfo != null)
                {
                    priceValue = ProductManager.ProductProvider.GetPrice(productInfo, quantity);
                }

                var price = priceValue == null
                    ? null
                    : new PriceRaw
                    {
                        Price = priceValue.Value,
                        Currency = currency
                    };

                return price;
            }
            catch (Exception e)
            {
                Logger.Instance.Log(ErrorLevel.Error, $"Unknown error during FindPrice(). Exception: {e.Message}");
                return null;
            }
        }

        private void RestoreVariantId(Product product, long variantProductAutoId, string productVariantId)
        {
            if (product.AutoId == variantProductAutoId && !string.Equals(product.VariantId, productVariantId))
            {
                product.VariantId = productVariantId;
            }
        }

        /// <summary>
        /// Prepares the prices.
        /// </summary>
        /// <param name="products">The products.</param>
        public override void PreparePrices(Dictionary<Product, double> products)
        {
            // This overload is invoked when OrderLines are prepared (only).
        }

        /// <summary>
        /// Prepares the prices.
        /// </summary>
        /// <param name="products">The products.</param>
        public override void PreparePrices(ProductCollection products)
        {
            // This overload is invoked when products are rendered.
            if (products != null && products.Count > 0
              && Connector.IsWebServiceConnectionAvailable()
              && Global.IsIntegrationActive
              && Global.IsIntegrationEnabledForShop)
            {
                ProductManager.FetchProductInfos(products.ToList());
            }
        }
    }
}