﻿using Dynamicweb.Ecommerce.Integration;
using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Ecommerce.LiveIntegration.Logging;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Ecommerce.Products;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml;

namespace Dynamicweb.Ecommerce.LiveIntegration.Products
{
    /// <summary>
    /// Default implementation of a product provider to get prices and other info for products from the ERP.
    /// Inherit this class and override one or more of its methods to alter the behavior.
    /// </summary>
    /// <example>
    /// <code description="Example" source="..\..\s\Dynamicweb\Integrations\Dynamicweb.Ecommerce.LiveIntegration\Head\Dynamicweb.Ecommerce.LiveIntegration.Examples\CustomProductProvider.cs" lang="CS"></code>
    /// </example>
    public class ProductProviderBase
    {
        /// <summary>
        /// Creates a unique product identifier by concatenating the product ID or number (depends on the CalculatePriceUsingProductNumber setting), the variant ID and the language ID.
        /// Override to build up your own unique identifier.
        /// </summary>
        /// <param name="product">The product.</param>
        /// <returns>System.String.</returns>
        public virtual string GetProductIdentifier(Product product)
        {
            return GetProductIdentifier(product, null);
        }

        /// <summary>
        /// Creates a unique product identifier by concatenating the product ID or number (depends on the CalculatePriceUsingProductNumber setting), the variant ID and the language ID.
        /// Override to build up your own unique identifier.
        /// </summary>
        /// <param name="product">The product.</param>
        /// <param name="unitId">The product unit Id.</param>
        /// <returns>System.String.</returns>
        public virtual string GetProductIdentifier(Product product, string unitId)
        {
            string unit = string.Empty;
            if (!string.IsNullOrEmpty(unitId) && Settings.Instance.UseUnitPrices)
            {
                unit = $".{unitId}";
            }
            if (Settings.Instance.CalculatePriceUsingProductNumber)
            {
                return $"{product.Number}.{product.VariantId}.{product.LanguageId}{unit}";
            }
            else
            {
                return $"{product.Id}.{product.VariantId}.{product.LanguageId}{unit}";
            }
        }

        /// <summary>
        /// Gets the price.
        /// </summary>
        /// <param name="product">The product.</param>
        /// <param name="quantity">The quantity.</param>
        /// <returns>System.Nullable&lt;System.Double&gt;.</returns>
        /// <exception cref="ArgumentNullException">product</exception>
        /// <example>
        /// <code description="Overriding example" source="..\..\s\Dynamicweb\Integrations\Dynamicweb.Ecommerce.LiveIntegration\Head\Dynamicweb.Ecommerce.LiveIntegration.Examples\CustomProductProvider.cs" lang="CS"></code>
        /// </example>
        public virtual double? GetPrice(ProductInfo product, double quantity)
        {
            if (product == null)
            {
                throw new ArgumentNullException(nameof(product));
            }

            double? price = (double?)product["TotalPrice"];

            if (!price.HasValue)
            {
                if (quantity <= 0)
                {
                    quantity = 1;
                }

                var erpPriceResponse = ((IList<ProductPrice>)product["Prices"] ?? Enumerable.Empty<ProductPrice>())
                       .Where(p => quantity >= p.Quantity.GetValueOrDefault(1))
                       .OrderByDescending(p => p.Quantity.GetValueOrDefault(1))
                       .FirstOrDefault();

                price = erpPriceResponse?.Amount;
            }

            return price;
        }

        /// <summary>
        /// Fills the product values.
        /// </summary>
        /// <param name="productInfo">The product information.</param>
        /// <param name="product">The product.</param>
        /// <param name="quantity">The quantity.</param>
        public virtual void FillProductValues(ProductInfo productInfo, Product product, double quantity)
        {
            double? newStock = (double?)productInfo["Stock"];
            if (newStock.HasValue && Helpers.IsDifferentStock(product.Stock, newStock.Value))
            {
                product.Stock = newStock.Value;
                Services.Products.Save(product);
                productInfo["Stock"] = null;
            }

            double? priceValue = GetPrice(productInfo, quantity);
            var productPrice = product.GetPrice(Common.Context.Currency.Code, Common.Context.Country.Code2); // NOTE accessing the price, will trigger the price provider to kick in
            productPrice.PriceWithoutVAT = priceValue.GetValueOrDefault(); 
            productPrice.PriceWithVAT = priceValue.GetValueOrDefault();

            // Update Product Custom Fields
            if (Settings.Instance.AddProductFieldsToRequest && product.ProductFieldValues.Count > 0)
            {
                foreach (ProductFieldValue pfv in product.ProductFieldValues)
                {
                    if (productInfo.ContainsKey(pfv.ProductField.SystemName))
                    {
                        pfv.Value = productInfo[pfv.ProductField.SystemName];
                    }
                }
            }
        }

        /// <summary>
        /// Gets the product prices list.
        /// </summary>        
        /// <param name="productInfo">The product information.</param>
        public virtual List<Price> GetProductPrices(ProductInfo productInfo)
        {
            var result = new List<Price>();
            var prices = (IList<ProductPrice>)productInfo["Prices"];

            foreach (var price in prices.OrderBy(p => p.Quantity.GetValueOrDefault()))
            {
                result.Add(new Price
                {
                    Id = price.Id,
                    Amount = price.Amount.GetValueOrDefault(),
                    Quantity = price.Quantity.GetValueOrDefault(),
                    ProductId = price.ProductId,
                    VariantId = price.ProductVariantId,
                    LanguageId = Common.Context.LanguageID,
                    UserCustomerNumber = price.UserCustomerNumber,
                    IsInformative = false,
                    CurrencyCode = Common.Context.Currency.Code,
                    UnitId = price.UnitId
                });
            }
            return result;
        }

        /// <summary>
        /// Extracts the prices.
        /// </summary>
        /// <param name="response">The response.</param>
        /// <returns>IList&lt;ProductPrice&gt;.</returns>
        public virtual IList<ProductPrice> ExtractPrices(XmlDocument response)
        {
            var prices = new List<ProductPrice>();

            var pricesXml = response.SelectNodes("//item [@table='EcomPrices']");

            if (pricesXml != null)
            {
                foreach (XmlNode priceXml in pricesXml)
                {
                    string customerNumber = priceXml.SelectSingleNode("column [@columnName='PriceUserCustomerNumber']")?.InnerText;
                    if (string.Equals(customerNumber, Settings.Instance.AnonymousUserKey, StringComparison.OrdinalIgnoreCase))
                    {
                        customerNumber = string.Empty;
                    }

                    var price = new ProductPrice
                    {
                        Id = priceXml.SelectSingleNode("column [@columnName='PriceId']")?.InnerText,
                        ProductId = priceXml.SelectSingleNode("column [@columnName='PriceProductId']")?.InnerText,
                        ProductVariantId = priceXml.SelectSingleNode("column [@columnName='PriceProductVariantId']")?.InnerText,
                        UserCustomerNumber = customerNumber,
                    };

                    var node = priceXml.SelectSingleNode("column [@columnName='PriceAmount']");
                    price.Amount = (node != null && !string.IsNullOrEmpty(node.InnerText)) ? Helpers.ToDouble(node.InnerText) : (double?)null;

                    node = priceXml.SelectSingleNode("column [@columnName='PriceQuantity']");
                    price.Quantity = (node != null && !string.IsNullOrEmpty(node.InnerText)) ? Helpers.ToDouble(node.InnerText) : (double?)null;

                    price.UnitId = priceXml.SelectSingleNode("column [@columnName='PriceProductUnitId']")?.InnerText;

                    prices.Add(price);
                }
            }

            return prices;
        }

        /// <summary>
        /// Checks if the product is enabled for the live prices requests.
        /// Override this method if you want to skip some products from being looked up in the ERP.
        /// </summary>
        /// <param name="product">The product</param>
        /// <returns>True if product is enabled for the live prices, otherwise false.</returns>
        public virtual bool IsLivePriceEnabledForProduct(Product product)
        {
            return true;
        }

        /// <summary>
        /// Gets the product from variant combo identifier.
        /// </summary>
        /// <param name="product">The product.</param>
        /// <returns>Product.</returns>
        public virtual Product GetProductFromVariantComboId(Product product)
        {
            try
            {
                // If master: exchange master product with default variant
                if (product.IsVariantMaster && product.DefaultVariantComboId.Length > 0)
                {
                    Product returnProduct = product;

                    string sql =
                        $"SELECT TOP 1 * FROM EcomProducts WHERE ProductId='{product.Id}' AND ProductVariantId='{product.DefaultVariantComboId}' AND ProductLanguageId='{product.LanguageId}'";
                    foreach (Product p in Product.GetProductBySql(sql))
                    {
                        returnProduct = p;
                    }

                    returnProduct.VariantId = string.Empty;
                    return returnProduct;
                }
                else
                {
                    product.VariantId = string.Empty;
                    return product;
                }
            }
            catch (Exception e)
            {
                Logger.Instance.Log(ErrorLevel.Error, $"Failed getting product by method GetProductFromVariantComboId: '{e.Message}'.");

                product.VariantId = string.Empty;
                return product;
            }
        }
    }
}